/* 
   Program to use the commands in 
   the GPIB--Driver gpib_drv without coding a program.
   
   syntax: ieee [Options]
                {command}
                if addressed command one or more devices or gpib0 will be taken
                if a parameter is needed it is the last argument on the command line
                         or - if possible - it is generated internally
                
 Written & (C) 1995,96 by Michael Boegl, Thomas Hohenberger
 Lehrstuhl Experimentalphysik V
 Universitaet Bayreuth
 e-mail: ieee@lowt.phy.uni-bayreuth.de
 
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation with the modifications stated in the 
   'COPYING' licence file.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   (See file 'GPL' distributed with this package.)

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/


#include <stdio.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <linux/errno.h>
#include <string.h>
#include <ctype.h>
		     
#include "gpib_drv.h"

#define READ_SIZE 1024   /* max length of files to read in for cmd WRITF */

#define AddComGro 1
#define UniComGro 2
#define ManLinCom 3
#define DriverCom 4
#define DevSetCom 5
#define DivUseCom 6

#define IEEE_READ 1
#define IEEE_WRIT 2
#define IEEE_WRITF 3
#define IEEE_QUERY 4

#define IEEE_VERSION "V0.91beta jul-96"

char *howto = "ieee - a programm to use with the gpib_drv - kernel module \n"
              "       written & (C) by Michael Boegl & Thomas Hohenberger, EPV, Uni Bayreuth\n"
              "       "IEEE_VERSION" for "GPIB_VERSION"\n"
              "\nieee and gpib_drv.o are coming with absolutely NO WARRANTY.\n"
              "This is free software. See COPYING file for licence information.\n"
              "\nusage: ieee [opt] {cmd} [dev [dev...]] [par]\n"
              "            opt: -v (verb. mode), -? (shows this), -l (cmd list)\n"
              "            cmd: a ieee command (see manual)\n"
              "            dev: name of devices, connected to ieee488 interface bus\n"
              "            par: if needed the last argument will be the parameter\n"
              "                                       (strings must be in quotes!)\n";

char *cmds = "addressed ieee488 commands:\n"
             " GTL\n"
             " SDC\n"
             " PPC\n"
             " GET\n"
             " TCT\n"
             "universal ieee488 commands:\n"
             " LLO\n"
             " DCL\n"
             " PPU\n"
             " SPE\n"
             " SPD\n"
             "ieee488 management-line controls:\n"
             " IFC\n"
             " SREN\n"
             " CREN\n"
             "gpib_drv kernel module commands:\n"
             " NUL\n"
             " ONL\n"
             " RST\n"
             " CMD\n"
             "device setting commands:\n"
             " EOF\n"
             " CRLF\n"
             " TIMEOUT\n"
             " SETEOS\n"
             " SHAREDEV\n"
             " SENDLASTBYTE\n"
#ifdef SLOW
             " SETDELAY\n"
#endif
             " ENDONEOS\n"
             "read/write commands:\n"
             " READ\n"
             " WRITE\n"
             " WRITEF\n"
             " QUERY\n";


char devname[256]="";
int  request=GPIB_NUL;        /* ieee-command (request for the ioctl() routine) */
unsigned long parameter=0L;   /* parameter to pass to ioctl()                   */
int anzpara=0;                /* number of parameters needed for ieee-command   */
int comgrp=0;                 /* command group (see defines above)              */
int verb=0;                   /* verbose -- mode                                */
int anzopt=0;                 /* number of options                              */
char errbuf[256];             /* buffer for error text                          */
char cmdstr[256];             /* contains the gpib_drv cmdstring                */
int returncode = 0;           /* program return code */

void exit_ieee(char* txt,int level)
{
 if ( (level <= 10) && (verb) ) fprintf(stderr,"WARNING: %d, %s\n",level,txt);
 if ( level >  10 ) fprintf(stderr,"ERROR: %d, %s\n",level,txt),exit (-level);
}


void usage(int i)
{
 fprintf(stderr,howto);
 if (i==2) fprintf(stderr,cmds);
 exit(0);
}


void checkoptions(int argc, char *argv[])
{
 int i;

 if (argc<2) usage(1);

 for (i=1;i<argc;i++)
   {
    if (!strcmp("-?",argv[i]))     usage(1);
    if (!strcmp("?",argv[i]))      usage(1);
    if (!strcmp("--help",argv[i])) usage(1);
    if (!strcmp("-l",argv[i]))     usage(2);
    if (!strcmp("-v",argv[i]))     anzopt++, verb=1;
   }
}


int checkerr(int dev)  /* Tom! */
{
    int err,sta;
    if (dev)
	{
	    err=ioctl(dev,GPIB_ERR);
	    sta=ioctl(dev,GPIB_STA);
	    if (verb) fprintf (stderr,"GPIB: error=%04x, status=%04x\n",err,sta);
	    if (err) fprintf (stderr,"WARNING: checkerr: A GPIB--error had occured: error=%04x\n",err);
	    return (0);
	}
    else
	exit_ieee ("checkerr: device not open",11);
}


int numadress()
{
 int adr=0,ret=0;

 adr = open (devname,O_RDWR|O_SYNC);
 if ( adr < 0 ) 
    ret=adr;
 else 
    {
     ret = ioctl(adr,GPIB_NUL);
     close (adr);
    }
 return (ret);
}


void mkdevname(char *str)
{
 if (str[0]!='/') strcpy(devname,GPIB_PATH), strcat(devname,str);
 else strcpy(devname,str);
 if ( numadress() < 0 ) strcpy (devname,"NULL");
}


void translate (char *str)
{
 int i;
 for (i=0;i<strlen(str);i++) *(str+i)=tolower(*(str+i));

   /* Addressed Command Group */ 
 comgrp = AddComGro;
 if (!strcmp("gtl",str)) { request = GPIB_GTL; return; }
 if (!strcmp("sdc",str)) { request = GPIB_SDC; return; }
 if (!strcmp("ppc",str)) { request = GPIB_PPC; return; }
 if (!strcmp("get",str)) { request = GPIB_GET; return; }
 if (!strcmp("tct",str)) { request = GPIB_TCT; return; }

   /* Universal Command Group */
 comgrp = UniComGro;
 if (!strcmp("llo",str)) { request = GPIB_LLO; return; }
 if (!strcmp("dcl",str)) { request = GPIB_DCL; return; }
 if (!strcmp("ppu",str)) { request = GPIB_PPU; return; }
 if (!strcmp("spe",str)) { request = GPIB_SPE; return; }
 if (!strcmp("spd",str)) { request = GPIB_SPD; return; }

   /* Management Line Commands */
 comgrp = ManLinCom;
 if (!strcmp("ifc",str))  { request = GPIB_IFC; return; }
 if (!strcmp("sren",str)) { request = GPIB_SREN; return; } 
 if (!strcmp("cren",str)) { request = GPIB_CREN; return; }

   /* Driver Commands */
 comgrp = DriverCom;
 if (!strcmp("nul",str)) { request = GPIB_NUL; return; }
 if (!strcmp("onl",str)) { request = GPIB_ONL; return; }
 if (!strcmp("rst",str)) { request = GPIB_RST; return; }
 if (!strcmp("cmd",str)) { request = GPIB_CMD; anzpara=1; return; }

   /*  Device Settings Commands */
 comgrp = DevSetCom;
 if (!strcmp("eof"     ,str)) { request = GPIB_DEV_EOF;      anzpara = 1; return; }
 if (!strcmp("crlf"    ,str)) { request = GPIB_DEV_CRLF;     anzpara = 1; return; }
 if (!strcmp("timeout" ,str)) { request = GPIB_DEV_TIMEOUT;  anzpara = 1; return; }
 if (!strcmp("seteos"  ,str)) { request = GPIB_DEV_SETEOS;   anzpara = 1; return; }
 if (!strcmp("sharedev",str)) { request = GPIB_DEV_SHAREDEV; anzpara = 1; return; }
#ifdef SLOW
 if (!strcmp("setdelay",str)) { request = GPIB_DEV_SETDELAY; anzpara = 1; return; }
#endif
 if (!strcmp("endoneos",str)) { request = GPIB_DEV_ENDONEOS; anzpara = 1; return; }
 if (!strcmp("sendlastbyte",str)) { request = GPIB_DEV_SENDLASTBYTE; anzpara = 1; return; }
   
   /* Diverse Useful Comands */
 comgrp = DivUseCom;
 if (!strcmp("read" ,str))  { request = IEEE_READ;  return; } 
 if (!strcmp("write",str))  { request = IEEE_WRIT;  return; } 
 if (!strcmp("writef",str)) { request = IEEE_WRITF; return; } 
 if (!strcmp("query",str))  { request = IEEE_QUERY; return; }
 
 sprintf(errbuf,"The command  %s  is not implemented",str);
 exit_ieee (errbuf,11);
}


/* 
 * do_command(argc,argv,firstdev,lastdev,altnativedev,once,para,req)
 *         firstdev: argv[firstdev]
 *         lastdev:  argv[lastdev]
 *         alternative: takes gpib0 if no device given
 *         once =1 breaks on success, once = 0 restarts with next dev-name
 *         para ist the parameter to send to ioctl
 *         req is the requested operation
 */
void do_command(int argc, char *argv[], int fdev, int ldev, int adev, int once, long para, int req)
{
 int i,add;


 for (i=fdev;i<=ldev;i++)   /* Es sind Devices angegeben */
    {
     mkdevname(argv[i]);
     if (strcmp("NULL",devname))
        {
         add = open (devname,O_RDWR|O_SYNC);
         if (add < 0) 
            sprintf(errbuf,"Cannot open %s, loc. 4a",devname),exit_ieee(errbuf,9);
         else
            {
             ioctl (add,req,para);
             close(add);
             if (once) return;       /* falls erfolgreich, abbrechen */
            }
	}
     else 
	{
	 sprintf(errbuf,"cannot open device %s",argv[i]);
	 if ( (i==ldev) && (!adev) ) exit_ieee(errbuf,13);
	 else exit_ieee(errbuf,5);
	}
    }
 if ( adev && (i==ldev+1) ) 
    {
     mkdevname("gpib0");
     if (strcmp("NULL",devname))
        {
         add = open (devname,O_RDWR|O_SYNC);
         if (add < 0) 
            sprintf(errbuf,"Cannot open %s, loc. 4a",devname),exit_ieee(errbuf,9);
         else
            {
             ioctl (add,request,parameter);
             close(add);
             return;
    	    }
	}
     else sprintf(errbuf,"cannot open device %s","gpib0"),exit_ieee(errbuf,12);
    }
 return;
}


void ieee_read (int argc, char *argv[], int akt, int anzbytes)
{
 int j,ii;
 int i=akt;
 char *buf;
 int valuegiven=1;

 mkdevname(argv[i++]);
 if (!strcmp("NULL",devname)) exit_ieee("cannot open device",13);
 if ((ii=open(devname,O_RDWR|O_SYNC))<0) exit_ieee("cannot open device",13);

 if ( anzbytes >  0 ) parameter = anzbytes;
 if ( (anzbytes < 0 ) && ( argc == i) ) parameter = -1*anzbytes;
 if ( (anzbytes < 0 ) && ( argc != i) ) parameter = atol(argv[i]);
 if ( anzbytes == 0 ) exit_ieee("cannot read in 0-bytes packages",19);
 if ((buf = (char *) malloc(parameter)) ==NULL) exit_ieee("not enough space",15);
/*  if (argc==i) valuegiven=0;  */
/*  else (parameter=atol(argv[i])); */
/*  if ((buf=(char *)malloc((valuegiven?parameter:(parameter=1024))))==NULL) exit_ieee("not enough space",15); */

 again:
 j=read(ii,buf,parameter);
 write (1,buf,j); 
 if (verb) fprintf(stderr,"GPIB: number of bytes read: %d\n",j);
 checkerr(ii);
 if ( (j==parameter) && !valuegiven ) goto again;
 if (ii) close(ii);
}


void ieee_writ (int argc, char *argv[], int akt)
{
 int j,ii;
 int i=akt;
 int jj,len;
 char *buf;
 
 mkdevname(argv[i++]);
 if (!strcmp("NULL",devname)) exit_ieee("cannot open device",13);
 if ((ii=open(devname,O_RDWR|O_SYNC))<0) exit_ieee("cannot open device",13);
 for (len=0,jj=i;jj<argc;++jj) len+=strlen(argv[jj])+1;
 if ((buf=(char *)malloc(len))==NULL) exit_ieee("not enough memory",15);
 *buf=0;
 for (jj=i;jj<argc;++jj) strcat(buf,argv[jj]), strcat(buf," ");
 buf[strlen(buf)-1]=0;
 j=write(ii,buf,strlen(buf));
 free (buf);
 if (verb) fprintf(stderr,"GPIB: number of bytes written: %d\n",j);
 checkerr(ii);
 if (ii) close(ii);
}


void ieee_writf (int argc, char *argv[], int akt)
{
 int j,ii;
 int filed=0;
 int i=akt;
 char *buf;
 char infilname[256];

 mkdevname(argv[i++]);
 if (!strcmp("NULL",devname)) exit_ieee("cannot open device",13);
 if ((ii=open(devname,O_RDWR|O_SYNC))<0) exit_ieee("cannot open device",13);
 if ((buf=(char *)malloc(READ_SIZE))==NULL) exit_ieee("not enough memory",15);
 if (argc!=i) 
   {
    strcpy(infilname,argv[i++]);
    filed=open(infilname,O_RDWR|O_SYNC);
   }
 do
   {
    i=read(filed,buf,READ_SIZE);
    j=write(ii,buf,i);
    if (verb) fprintf(stderr,"GPIB: number of bytes written: %d\n",j);
    checkerr(ii);
   } while (i==READ_SIZE);
 if (ii) close(ii);
 if (filed) close (filed);
}

int main (int argc, char *argv[])
{
 int i=1,j=1,ii;
 char la;
 char *buf;

 checkoptions(argc,argv);
 i+=anzopt;

 translate(argv[i++]); 
 
 switch (comgrp)
   {
    case AddComGro:
          if (argc < 3+anzopt)  /* something must be wrong */
	      sprintf (errbuf,"Addressed Command without Device"), exit_ieee(errbuf,5);
	  if (argc == 3+anzopt) /* only one device given */
	     do_command(argc,argv,i,argc-1,0,1,parameter,request);
          if (argc > 3+anzopt)	 /* make a command-string and go to ioctl(GPIB_CMD) */
             {
exit_ieee("addressed commands with more than one dev. not yet implemented",11);
              strcpy(cmdstr,UNL UNT);
              for (ii=i,j=2;ii<argc;ii++,j++)
		  {
		   mkdevname(argv[ii]);
		   la=0x20+numadress();
                   if (!strcmp(devname,"NULL")) j--;
		   else cmdstr[j]=la;
		  }
	      cmdstr[j++]=request;
	      cmdstr[j++]=0;
              do_command(argc,argv,i,argc-1,1,1,(long)cmdstr,GPIB_CMD);
	     }
	  break;
    case UniComGro:
    case ManLinCom: 
          do_command(argc,argv,i,argc-1,1,1,parameter,request);
	  break;
    case DriverCom:
          if ( anzpara == 1 )
	      parameter = (long)(argv[argc-1]);   /* last arg is command string */
          do_command(argc,argv,i,argc-1-anzpara,1,1,parameter,request);
	  break;
    case DevSetCom:
	  if ( anzpara == 1 )
	      parameter = atol(argv[argc-1]); /* convert last arg to parameter */
          do_command(argc,argv,i,argc-1-anzpara,0,0,parameter,request);
	  break;
    case DivUseCom:
	  switch (request)
	     {
	      case IEEE_READ:  /* read n bytes if n given, otherwise in 1024byte packets */
                  ieee_read (argc,argv,i,-1024);
		  break;
	      case IEEE_WRIT: /* write last arg (string) to dev */
                  ieee_writ (argc,argv,i);
                  break;
	      case IEEE_WRITF: /* write to dev, if last arg is file than this file, otherwise stdin */
		  ieee_writf (argc,argv,i);
		  break;
	      case IEEE_QUERY: /* write last arg(s) to dev, then read from dev as 1024 byte packages */
		  ieee_writ (argc,argv,i);
                  ieee_read (argc,argv,i,1024);
		  break;
              default: exit_ieee("How can you get here?????\n### Please contact bogi###\n",20);
	     }
    default: 
    }

 exit(returncode);
}


